/* this is a beta release for a can sequencer for prius gen3 on mbed
it is based on a work from:
Moon Light on mbed.org for the original idea for the sequencer
Ym1784 for the filtering && idea for ring buffer
2009Prius for external parameters


hardware:
    mbed LPC1768
    Can transceiver mcp2551 on pin 30 & 29
    BT module               on pin 28 & 27
    SD card                 on pin 5 6 7 8
    UART  LCD/GPS           on pin
    BMP                     on pin
main mods:


mod 2/04/2012: change detection for last frame in case of multiframe
mod 9/04/2012:: use external files on /local/ for filters & requests  * Author: teamprii

mod 10/04/2012 : use buffer for can_msg : ym1784
mod 10/04/2012 : change detection for end of multiframe,
    adapt position for positionning multiframe's sequence because of non sequential arrival...
mod 21/04/2012 : do calcs & broadcast results


priusfan @ priusfan.info
*/
//===============================================================
// the following parms are updated in the init sequence
//===============================================================
// log level
int loglvl=2; // 0= debug , 1= detail , 2 = synthesis

// hardware & channels
bool use_usb =true;
bool use_sdcard =false;
bool use_bt = false;
bool use_bmp = false;

bool use_requests = false;
bool use_filter = false;

//===============================================================


#include "mbed.h"
#include "SDHCFileSystem.h"
#include "stdarg.h"
#include "common.h"
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <bitset>
using namespace std;

LocalFileSystem local("local");
SDFileSystem sd(p5, p6, p7, p8, "sd"); // mosi, miso, sclk, cs
vector<reqTbl> reqTable;
int reqNumTbl;

Timer           timer;
DigitalOut      led1(LED1);
DigitalOut      led2(LED2);
DigitalOut      led3(LED3);
DigitalOut      led4(LED4);

// We use can on mbed pins 29(CAN_TXD) and 30(CAN_RXD).
CAN can2(p30, p29);

//Virtual serial port over USB
Serial pc(USBTX, USBRX); // tx, rx
//BT
Serial bt(p28,p27); // tx, rx

int   len_max = 0;

// this request will be completely modified according external file /local/request.txt
char  msg_Request[8]  ;
char  msg_Continue[8] = { 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }; // Continue Receiving Additional Data
int   msg_ReqIdx    = 0;
int   msg_ReqFrameW = 0;

int w_id;
int w_len;
char w_data[8];

//
char  msg_MF_data[100]       ; // used to store Multiframe
unsigned int msg_MF_Id       ; // Multiframe Id

char Crc                    ;   // crc char for checking integrity
char strtosnd[256], *str=strtosnd;
CANMessage can_MsgRx;
bool sdopen=false;
bool btopen=false;
bool sleeping=true;
bool sndtime=false;
Ticker wd;
//


//============================================
// declaration functions (prototypes)
//============================================
void init();
void CAN2_wrFilter (uint32_t id);
void can_check(void);
void wrkmsg();
void calc_g2(int id, char wrk[]);
void calc_g3(int id, char wrk[]);
void watchdog();
void dispatcher();

#define BUFFER_SIZE 128                     // buffer size must be power of 2
#define BUFFER_MASK ( BUFFER_SIZE - 1)
volatile unsigned int rb_write_pointer = 0; // ring buffer write pointer
volatile unsigned int rb_read_pointer = 0;  // ring buffer read pointer
volatile unsigned int rb_data_size = 0;     // ring buffer number of entries
unsigned int second[BUFFER_SIZE];
unsigned short can_id[BUFFER_SIZE];
unsigned char msg_length[BUFFER_SIZE];
unsigned char data[BUFFER_SIZE][8];
int wdc; // watchdog counter
int fps; // watchdog counter
FILE *fp;
//===========================

//===========================
std::map <int,char> id_mask; // map for association mask to an Id


//void calcs();
//void calcm();
//=======================================

int          send_per_sec =  0;
int          rec_per_sec =  0;

//============================================

int main() {
    init();

    while (1) {
        // read CAN messages from the buffer
        if ( rb_data_size > 0 ) {               // new message is available
            rb_read_pointer++;                      // position to next
            rb_read_pointer &= BUFFER_MASK;         // loop if at end of array
            rb_data_size--;                         // update nbr of entries in the ring buffer
            sleeping=false;
            if (loglvl == 0) {                       // log detail of received msg
                pc.printf("debug:%d,", rb_read_pointer);
                pc.printf("%d,", rb_write_pointer);
                pc.printf("%d,", rb_data_size);
                pc.printf("%08X,", second[rb_read_pointer]);
                pc.printf("t%03X%d", can_id[rb_read_pointer], msg_length[rb_read_pointer]);
                for (char i=0; i<msg_length[rb_read_pointer]; i++) {
                    pc.printf("%02X", data[rb_read_pointer][i]);
                } // for
                pc.printf("\r\n");
            } // end (loglvl = 0)
            // we store in w_vars the results : it is not funny to manage all these [][]
            w_id=can_id[rb_read_pointer];
            w_len=msg_length[rb_read_pointer];
            for (int i=0; i<w_len; i++) {
                w_data[i]=data[rb_read_pointer][i];
            } // for
            led4=!led4;
            wrkmsg();
        } //        end // if ( rb_data_size > 0 )
        if (sleeping == true) {         // energy saving
            wait (0.5);
            led1=!led1;
            led2=0;
            led3=0;
            led4=0;
        } //        end  if sleeping
    } // while
} // main


//*----------------------------------------------------------------------------
//  init()
//  description: read parm file on local
//  get loglevel
//  get hdw used
//  get & install filters
//  store requests
//*----------------------------------------------------------------------------
void init() {
    pc.baud(115200);
    bt.baud(115200);
    timer.reset();
    const int maxlinelen=128; // max line length in canhsd.txt file; 128 should be plenty for things like "0B4 // comment for 0B4 ....... " etc.
    char str[maxlinelen];
    string param;
    int id;
    int timask;
    vector<reqTbl> reqtable;
    ifstream infile("/local/canhsd.txt"); // hard coded pathname for parameters
    if (infile.is_open()) {
        while (infile.good()) {
            infile.getline(str,maxlinelen);
            stringstream ss;
            ss<<str;
            if (str[0]!='/' && ss.str().length() > 4 ) {   // ignore comments & short lines

                ss>> param;                         // get first string in the line

                if (param== "TIME") {
                    long dattime;
                    ss>>dattime;
                    //  set_time(1335425872);  // Set RTC time to 26/04/2012 @ 07:37:52
                    set_time(dattime);
                } // end  if param== ('TIME')

                if (param== "PC_BAUD ") {
                    double pcbaud;
                    ss>>pcbaud;
                    pc.baud(pcbaud);
                } // end param== ('PC_BAUD ')

                if (param== "LOGLEVEL") {
                    ss>>loglvl;
                } // end param== ('LOGLEVEL')

                if (param== "USE_BT") {
                    ss>>use_bt;
                    btopen=true;
                } // end param== ('USE_BT')

                if (param== "USE_SD") {
                    ss>>use_sdcard;
                } //   end param== ('USE_SD')

                if (param== "FILTER") {             //  Author: teamprii
                    use_filter=true;
                    timask=0xFF;                    // initialize mask in case of...
                    ss>>hex>>id>>hex>>timask;       // analyse the line, extract id & mask
                    char tcmask=timask;             // convert the mask from int to char
                    id_mask[id]=tcmask;             // store in the map the key: id and the value: mask
                    CAN2_wrFilter (id);             // install filter
                }  // end  param== ('FILTER')

                if (param== "REQUEST") {            //  Author: teamprii
                    use_requests=true;
                    int id; // PID
                    int freq; // request frequency
                    int codej; // to store read code temporarily
                    reqTbl tmp;
                    ss>>hex>>id>>dec>>freq;
                    // pc.printf("\r\n add this request to reqTable.\r\n");
                    tmp.id=id;
                    tmp.period=freq;
                    vector<int> code; // list of code
                    do {
                        ss>>hex>>codej;
                        if (ss.good()) {
                            code.push_back(codej);
                        }
                    } while (ss.good());

                    tmp.code=code;

                    if (loglvl==0) {
                        pc.printf("PID=%03X",tmp.id);
                        pc.printf("\tfreq=%03d",tmp.period);
                        pc.printf("\tCode=");
                        for (int j=0; j<(int)tmp.code.size(); j++) {
                            pc.printf(" %02x",tmp.code[j]);
                        } // end for
                        pc.printf("\r\n");
                    } // end (loglvl=0)
                    reqtable.push_back(tmp);
                } // end  if (param== "REQUEST"))
            }                                      // end if (str[0]!='/' && ss.str().length() > 4 )
        }                                          // end  while (infile.good())
        infile.close();
    }                                              // (infile.is_open())
    timer.start();
    can2.attach(&can_check);
    can2.frequency(500000);

    wd.attach(&watchdog, 1.0);      // watchdog called every second
    led1=0;
    led2=0;
    led3=0;
    led4=0;
    msg_ReqFrameW = 2;
}
//
// manage received can msg
//
void wrkmsg() {
    static int NbB, NbC,Pos;
    if (w_id < 0x700) {   // passive  frame ,
        msg_ReqFrameW--;
        wdc++;      // feed the watchdogcounter
        char tmask=id_mask[w_id];   // trying to get the mask from the map id_mask using w_id as key...
        int imask = tmask;          // convert the mask  from char to integer

        bitset<8> mask(imask);      // trick fom 2009Prius ; if it works, he will get Champagne
        if (loglvl < 2) {
            pc.printf("%03X",w_id);

            for (int i=0; i<w_len ; i++) {      // loop  from 0 to 7
                if (mask[7-i]) {              // interesting Y/N
                    pc.printf("%02X",w_data[i]);
                } // if (mask[i])
            } // for
            pc.printf("\r\n");
        } // end (loglvl < 2)
        calc_g3( w_id, w_data);

        // send next req
        if (msg_ReqFrameW <= 0 && use_requests == true) {
            // it is ok to send another request
            // because  we received at least X passive frames since completion of previous request
            // get the full message to send
            for (int i=0; i<8; i++) {
                msg_Request[i] = reqTable[msg_ReqIdx].code[i];
            }
            if (loglvl == 0) {
                // log request
                pc.printf("T\t%03X\t",reqTable[msg_ReqIdx].id);
                for (int i=0; i<8; i++) {
                    pc.printf(" ");
                    pc.printf("%02X", msg_Request[i]);
                } // for
                pc.printf("\r\n");
                // end  log request
            } // end (loglvl = 0)
            if (can2.write(CANMessage(reqTable[msg_ReqIdx].id, msg_Request,8))) {
                send_per_sec++;
                msg_ReqFrameW = 100;

                // next req, find msg_ReqIdx
                do {
                    reqTable[msg_ReqIdx].fcnt++;
                    if (reqTable[msg_ReqIdx].fcnt >= reqTable[msg_ReqIdx].period) {
                        reqTable[msg_ReqIdx].fcnt = 0;
                    } // endif
                    msg_ReqIdx++;
                    if (msg_ReqIdx >= reqNumTbl) {
                        msg_ReqIdx = 0;
                    } // endif
                } while (reqTable[msg_ReqIdx].fcnt != 0);
            } // end if can2.write
        } // end  if (msg_ReqFrameW <= 0)

    } else if (w_id >= 0x7E8 ) {        // from here, we are dealing with answers
        led2 =!led2;
        int HiNi = w_data[0]>>4;        // test HighNibble to detect continuation frame

        switch (HiNi) {
            case 0: {                      // short response  if (can_MsgRx.data[0] < 0x10) {

                if (loglvl < 2) {
                    pc.printf("%03X",w_id);
                    for (int i=0; i<8; i++) {
                        pc.printf("%02X",w_data[i]);
                    } // for
                    pc.printf("\r\n");
                } // end (loglvl < 2)
                calc_g3( w_id, w_data);
                msg_ReqFrameW = 2;
            } // case 0
            break;

            case 1 : {       // long response   if (can_MsgRx.data[0] == 0x10) {
                NbB =int(w_data[1]); // the Number of Bytes in the multiframe
                msg_ReqFrameW = 50;                  // continue receive
                // store id
                msg_MF_Id = w_id;
                // store from 1 to 7
                NbC = 0;
                for (int i = 1;  i < w_len ; i++) {
                    msg_MF_data[NbC] = w_data[i];
                    NbC ++;
                } // for
                // send contination request
                if (can2.write(CANMessage(w_id -8, msg_Continue, 8))) {
                    led3 =!led3;
                } // if
            } // case 1
            break;

            case 2 : { //               if (can_MsgRx.data[0] > 0x20 && can_MsgRx.data[0] <= 0x2F)
                int LoNi = w_data[0] & 0x0F;     // store low nibble, keep right part
                // store from 1 to 7 @ the right position according the low nibble
                Pos=   LoNi * 7;
                for (int i = 1; i < w_len ; i++) {
                    msg_MF_data[Pos + i - 1] = w_data[i];
                    NbC ++;
                } // for
                //
                if (NbC > NbB ) {   // we received the full multiframe , maybe in disorder, but the count is OK
                    msg_ReqFrameW = 2;
                    // multiframe is finished, we send the complete thing to pc
                    if (loglvl < 2) {
                        pc.printf("%03X",msg_MF_Id);
                        for (int i=0; i <=  NbB; i++) {
                            pc.printf("%02X", msg_MF_data[i]);
                        } // for
                        pc.printf("\r\n");
                    } // end (loglvl < 2)
                    calc_g3( msg_MF_Id, msg_MF_data);
                    // calcm();
                } // end  if (NbC > NbB )
            } // case 2
            // ignore
        } // end  switch (HiNi)
    } // end  if (w_id
}
//
// watchdog called every second
//
void watchdog () {
    sndtime=true;
    if (use_sdcard) {
        if ( sdopen && wdc < 10) { // if frame counter < 10 , close file on sdcard
            fclose(fp);
            sdopen=false;
        } // end if ( sdopen && wdc < 10)

        if ( !sdopen && (wdc > 40)) {   // if restart, open file on sdcard
            char file_name[30];
            time_t t;
            t = time( NULL );
            strftime( file_name, 20, "/sd/%m%d%H%M.log", localtime( &t ) );
            fp = fopen(file_name, "w");
            if (fp == NULL) {
                pc.printf("    Could not open file for write\r\n");
                sdopen=false;
            }  else {
                pc.printf(" open file for write%s\r\n",file_name);
                sdopen=true;
            }                                   //     end (fp == NULL)
        }                                       //     end ( !sdopen && (wdc > 40
    }                                           //     end (use_sdcard)

    fps=wdc;
    if (wdc < 10) {
        sleeping = true;
    }
    wdc=0;
}

/*----------------------------------------------------------------------------
  CAN_wrFilter()
 description: setup acceptance filter.  CAN controller (1..2)

  setup acceptance filter for CAN controller 2
  original http://www.dragonwake.com/download/LPC1768/Example/CAN/CAN.c
  simplified for CAN2 interface and std id (11 bit) only by YM1784
 *----------------------------------------------------------------------------*/
void CAN2_wrFilter (uint32_t id)  {
    static int CAN_std_cnt = 0;
    uint32_t buf0, buf1;
    int cnt1, cnt2, bound1;

    /* Acceptance Filter Memory full */
    if (((CAN_std_cnt + 1) >> 1) >= 512)
        return;                                       /* error: objects full */

    /* Setup Acceptance Filter Configuration
      Acceptance Filter Mode Register = Off  */
    LPC_CANAF->AFMR = 0x00000001;

    id |= 1 << 13;                        /* Add controller number(2) */
    id &= 0x0000F7FF;                            /* Mask out 16-bits of ID */

    if (CAN_std_cnt == 0)  {                     /* For entering first  ID */
        LPC_CANAF_RAM->mask[0] = 0x0000FFFF | (id << 16);
    }  else if (CAN_std_cnt == 1)  {             /* For entering second ID */
        if ((LPC_CANAF_RAM->mask[0] >> 16) > id)
            LPC_CANAF_RAM->mask[0] = (LPC_CANAF_RAM->mask[0] >> 16) | (id << 16);
        else
            LPC_CANAF_RAM->mask[0] = (LPC_CANAF_RAM->mask[0] & 0xFFFF0000) | id;
    }  else  {
        /* Find where to insert new ID */
        cnt1 = 0;
        cnt2 = CAN_std_cnt;
        bound1 = (CAN_std_cnt - 1) >> 1;
        while (cnt1 <= bound1)  {                  /* Loop through standard existing IDs */
            if ((LPC_CANAF_RAM->mask[cnt1] >> 16) > id)  {
                cnt2 = cnt1 * 2;
                break;
            }
            if ((LPC_CANAF_RAM->mask[cnt1] & 0x0000FFFF) > id)  {
                cnt2 = cnt1 * 2 + 1;
                break;
            }
            cnt1++;                                  /* cnt1 = U32 where to insert new ID */
        }                                          /* cnt2 = U16 where to insert new ID */

        if (cnt1 > bound1)  {                      /* Adding ID as last entry */
            if ((CAN_std_cnt & 0x0001) == 0)         /* Even number of IDs exists */
                LPC_CANAF_RAM->mask[cnt1]  = 0x0000FFFF | (id << 16);
            else                                     /* Odd  number of IDs exists */
                LPC_CANAF_RAM->mask[cnt1]  = (LPC_CANAF_RAM->mask[cnt1] & 0xFFFF0000) | id;
        }  else  {
            buf0 = LPC_CANAF_RAM->mask[cnt1];        /* Remember current entry */
            if ((cnt2 & 0x0001) == 0)                /* Insert new mask to even address */
                buf1 = (id << 16) | (buf0 >> 16);
            else                                     /* Insert new mask to odd  address */
                buf1 = (buf0 & 0xFFFF0000) | id;

            LPC_CANAF_RAM->mask[cnt1] = buf1;        /* Insert mask */

            bound1 = CAN_std_cnt >> 1;
            /* Move all remaining standard mask entries one place up */
            while (cnt1 < bound1)  {
                cnt1++;
                buf1  = LPC_CANAF_RAM->mask[cnt1];
                LPC_CANAF_RAM->mask[cnt1] = (buf1 >> 16) | (buf0 << 16);
                buf0  = buf1;
            }

            if ((CAN_std_cnt & 0x0001) == 0)         /* Even number of IDs exists */
                LPC_CANAF_RAM->mask[cnt1] = (LPC_CANAF_RAM->mask[cnt1] & 0xFFFF0000) | (0x0000FFFF);
        }
    }
    CAN_std_cnt++;

    /* Calculate std ID start address (buf0) and ext ID start address <- none (buf1) */
    buf0 = ((CAN_std_cnt + 1) >> 1) << 2;
    buf1 = buf0;

    /* Setup acceptance filter pointers */
    LPC_CANAF->SFF_sa     = 0;
    LPC_CANAF->SFF_GRP_sa = buf0;
    LPC_CANAF->EFF_sa     = buf0;
    LPC_CANAF->EFF_GRP_sa = buf1;
    LPC_CANAF->ENDofTable = buf1;

    LPC_CANAF->AFMR = 0x00000000;                  /* Use acceptance filter */
} // CAN2_wrFilter





/* =================================================
// CAN2 read using interrupt and ring buffer
 Author:    ym1784
 modified:  priusfan
=====================================================*/
void can_check(void) {
    while ( can2.read(can_MsgRx) ) {

        rb_write_pointer++ ;                //  increase write_pointer
        rb_write_pointer &= BUFFER_MASK;    //  loop if at end of array
        rb_data_size++ ;                    //  ring buffer number of entries

        if ( rb_write_pointer == rb_read_pointer && rb_data_size == BUFFER_SIZE) {  // buffer overflow
            fprintf(stderr, "buffer overflow\r\n");
            fprintf(stderr, "head = %d\r\n", rb_write_pointer);
            fprintf(stderr, "tail = %d\r\n", rb_read_pointer);
        }       // error
        // add entries in the ring buffer
        second[rb_write_pointer] = timer.read_us();
        can_id[rb_write_pointer] = can_MsgRx.id & 0xFFF;
        msg_length[rb_write_pointer] = can_MsgRx.len;
        for (char i=0; i<can_MsgRx.len; i++) {
            data[rb_write_pointer][i] = can_MsgRx.data[i];
        } // for
        led2 = !led2; // toggle led2
    } // while
} // can_check

//============================================
// calc_g3(int id, char wrk[])
//  do calcs & broadcast synthetic results
//  Author:    priusfan
//============================================


void calc_g3(int  id , char  wrk[]) {
// var declaration
    static int mg1_t, mg1_rpm, mg1_tor;
    static int mg2_t, mg2_rpm, mg2_tor;
    static int ice_t, ice_rpm, ice_tor;
    static int bcu_t, bcl_t;
    static int vlb, vlh;
    static int batt_ti, batt_t1, batt_t2, batt_t3;
    static int batt_dcl, batt_ccl;
    static int cat_t1, cat_t2;
    static int inv1_t, inv2_t;
    static double bat_soc, bat_amp;
    static double speed;

    switch (id) {
        case 0x00B4 : {
            speed=(wrk[5]*256+wrk[7])/100;
        }
        break;
        case 0x01C4 : {
            ice_rpm=(wrk[0]*256+wrk[1])/1.28;
        }
        break;
        case 0x0245 : {
            int acc_ped=wrk[2]/2;
        }
        break;
        case 0x03D3 : {
            double conso=(wrk[0]*256+wrk[1])/256;
        }
        break;
        case 0x0498 : {
        }
        break;
        case 0x04A6 : {
            int fuel_lvl=wrk[2]/2;
        }
        break;
        case 0x07E8 : {
            // inj
            if (wrk[1]==0x61 && wrk[2]==0x3C) {
            }
            // standard OBD http://en.wikipedia.org/wiki/OBD-II_PIDs
            if (wrk[1]==0x41 && wrk[2]==0x06) {
                double stft=  (wrk[3] -128) * 100 /128; // Short Term Fuel Trim     01 06   (A-128) * 100/128
                double ltft=  (wrk[5] -128) * 100 /128; // Long Term Fuel Trim      01 07   (A-128) * 100/128
                int ign= wrk[7] / 2 - 64;               // Timing Advance           01 0E   A/2 - 64
                int n=sprintf(str,"<FTI\t%3.2f\t%3.2f\t%d\t>\r\n",stft,ltft,ign);
                dispatcher();

            }
            // ice torque
            if (wrk[1]==0x61 && wrk[2]==0x49) {
                ice_tor= wrk[6] *256 + wrk[7] -32768;

                int n=sprintf(str, "<ICE\t%d\t%d\t%d\t%3.1f\t>\r\n",ice_rpm,ice_tor,ice_t,speed);
                dispatcher();
            }
            // cat temps
            if (wrk[1]==0x41 && wrk[2]==0x3C) {
                cat_t1=(wrk[3]*256+wrk[4])/10-40;
                cat_t2=(wrk[6]*256+wrk[7])/10-40;

                int n=sprintf(str,"<CATT\t%d\t%d\t>\r\n",cat_t1,cat_t2);
                dispatcher();
            }
        }
        break;
        case 0x07EA : {
            //MG1 rpm, torque, temp
            if (wrk[1]==0x61 && wrk[2]==0x61) {
                mg1_t=wrk[3]-40;
                mg1_rpm=wrk[6]*256+wrk[7]-32768;
                mg1_tor=(wrk[9]*256+wrk[10])/8 - 4096;

                int n=sprintf(str,"<MG1\t%d\t%d\t%d\t>\r\n",mg1_rpm,mg1_tor,mg1_t);
                dispatcher();
            }
            //MG2 rpm, torque, temp
            if (wrk[1]==0x61 && wrk[2]==0x62) {
                mg2_t=wrk[3]-40;
                mg2_rpm=wrk[6]*256+wrk[7]-32768;
                mg2_tor=(wrk[9]*256+wrk[10])/8 - 4096;

                int n=sprintf(str,"<MG2\t%d\t%d\t%d\t>\r\n",mg2_rpm,mg2_tor,mg2_t);
                dispatcher();
            }
            //Amp
            if (wrk[1]==0x61 && wrk[2]==0x8A) {
                bat_amp=(wrk[3]*256+wrk[4])/100 -327.68;
                bcu_t=wrk[6]-40;
                bcl_t=wrk[7]-40;
                vlb=(wrk[11]*256+wrk[12])/2;
                vlh=(wrk[13]*256+wrk[14])/2;

                int n=sprintf(str,"<BATT\t%3.2f\t%2.2f\t%d\t%d\t%d\t%d\t%d\t%d\t>\r\n",bat_amp,bat_soc,vlb,vlh,bcu_t,bcl_t,inv1_t,inv2_t);
                dispatcher();
            }
            //SoC
            if (wrk[1]==0x61 && wrk[2]==0x01) {
                bat_soc = wrk[24]/2.55;
                ice_t=wrk[8]-40;

            }
            //temps bc HVL HVH now combined with 21 8A
            /*      if (wrk[1]==0x61 && wrk[2]==0x74) {
                      bcu_t=wrk[3]-40;
                      bcl_t=wrk[4]-40;
                      vlb=(wrk[8]*256+wrk[9])/2;
                      vlh=(wrk[10]*256+wrk[11])/2;
                  } */
            //Inverters Temps
            if (wrk[1]==0x61 && wrk[2]==0x70) {
                inv1_t=wrk[3]-40;
                inv2_t=wrk[8]-40;
            }
            //Batt temps
            if (wrk[1]==0x61 && wrk[2]==0x87) {
                batt_ti=(wrk[3]*256+wrk[4])/256-50;
                batt_t1=(wrk[5]*256+wrk[6])/256-50;
                batt_t2=(wrk[7]*256+wrk[8])/256-50;
                batt_t3=(wrk[9]*256+wrk[10])/256-50;

                int n=sprintf(str,"<TBAT\t%d\t%d\t%d\t%d\t>\r\n",batt_ti,batt_t1,batt_t2,batt_t3);
                dispatcher();
            }
            //DCL CCL
            if (wrk[1]==0x61 && wrk[2]==0x98) {
                batt_dcl=wrk[9];
                batt_ccl=wrk[10];

                int n=sprintf(str,"<CDM\t%d\t%d\t>\r\n",batt_dcl,batt_ccl);
                dispatcher();
            }
        }
        break;
    }
}

// common message dispatcher
// according the hdw, it will send str on the right chanel
// also this is the place for timestamping (to avoid breaking messages)
void dispatcher() {

    if (sndtime==true && use_sdcard && sdopen ) {
        time_t seconds = time(NULL);
        char buffer[10];
        strftime(buffer, 10, "%X", localtime(&seconds));
        fprintf(fp,"<TIME\t%s\t%d>\r\n",buffer,wdc);
        sndtime=false;
    }

    if (use_usb) {
        pc.printf("%s",str);
    }

    if (use_sdcard && sdopen) {
        fprintf(fp,"%s",str);
    }

    if (use_bt && btopen) {
        bt.printf("%s",str);
    }

}